-- read a named rr file and return the results

CLASS_IN,CLASS_ANY = 1,255
TYPE_A,TYPE_NS,TYPE_CNAME,TYPE_SOA,TYPE_PTR,TYPE_MX,TYPE_TXT,TYPE_AAAA,TYPE_SPF = 1,2,5,6,12,15,16,28,99

function rr_type(t)
  if not t then return end
  t = t:lower()
  if t == 'a'     then return TYPE_A     end
  if t == 'ns'    then return TYPE_NS    end
  if t == 'cname' then return TYPE_CNAME end
  if t == 'soa'   then return TYPE_SOA   end
  if t == 'ptr'   then return TYPE_PTR   end
  if t == 'mx'    then return TYPE_MX    end
  if t == 'txt'   then return TYPE_TXT   end
  if t == 'aaaa'  then return TYPE_AAAA  end
  if t == 'spf'   then return TYPE_SPF   end
end

function rr_class(c)
  if not c then return end
  c = c:lower()
  if c == 'in'  then return CLASS_IN  end
  if c == 'any' then return CLASS_ANY end
end

function rr_time(t)
  if not t or #t == 0 then return end
  local total,part = 0
  for i=1,#t do
    local c = t:sub(i,i):lower()
    if c == 's' then
      if not part then return end
      total,part = total + part
    elseif c == 'm' then
      if not part then return end
      total,part = total + 60 * part
    elseif c == 'h' then
      if not part then return end
      total,part = total + 3600 * part
    elseif c == 'd' then
      if not part then return end
      total,part = total + 86400 * part
    elseif c == 'w' then
      if not part then return end
      total,part = total + 604800 * part
    else
      if not tonumber(c) then return end
      if not part then
        part = c
      else
        part = part * 10 + c
      end
    end
  end
  if not part then return total end
  return total + part
end

function parse_file(filename, origin)
  local fp = io.open(filename, 'r')
  if not fp then return end
  local text = fp:read('*a')
  local lines,c,token,tokens,in_parens,in_quote,in_comment = {},'','',{},0,false,false
  for i=1,#text do
    if c == '\\' then
      token = token .. text:sub(i,i)
    else
      c = text:sub(i,i)
      if in_comment then
        if c == '\r' or c == '\n' then in_comment = false end
      elseif in_quote then
        if c == '"' then
          in_quote = false
          if token ~= '' then tokens[#tokens+1],token = token,'' end
        else
          token = token .. text:sub(i,i)
        end
      else
        if c == ' ' or c == '\t' then
          if token ~= '' then tokens[#tokens+1],token = token,'' end
        elseif c == '(' then
          if token ~= '' then tokens[#tokens+1],token = token,'' end
          in_parens = in_parens + 1
        elseif c == ')' then
          if token ~= '' then tokens[#tokens+1],token = token,'' end
          if in_parens > 0 then in_parens = in_parens - 1 end
        elseif c == '"' then
          in_quote = true
        elseif c == ';' then
          if token ~= '' then tokens[#tokens+1],token = token,'' end
          if in_parens == 0 and #tokens ~= 0 then lines[#lines+1],tokens = tokens,{} end
          in_comment = true
        elseif c == '\r' or c == '\n' then
          if token ~= '' then tokens[#tokens+1],token = token,'' end
          if in_parens == 0 and #tokens ~= 0 then lines[#lines+1],tokens = tokens,{} end
        elseif c == '\\' then
        else
          token = token .. text:sub(i,i)
        end
      end
    end
  end
  if token ~= '' then tokens[#tokens+1],token = token,'' end
  if #tokens ~= 0 then lines[#lines+1],tokens = tokens,{} end

  local ret,name,ttl,type,class = {},origin,0
  for i=1,#lines do
    if lines[i][1] == '$ORIGIN' then
      origin = lines[i][2]
    elseif lines[i][1] == '$INCLUDE' then

    elseif lines[i][1] == '$TTL' then
      ttl = rr_time(lines[i][2])
    else
      if lines[i][1] == '@' then lines[i][1] = origin end
      local type_index
      for j = 1,4 do
        local t = rr_type(lines[i][j])
        if t then
          type = t
          type_index = j - 1
          break
        end
      end
      if type_index then
        for j=1,type_index do
          local c = rr_class(lines[i][j])
          if c then
            class = c
            lines[i][j] = nil
            break
          end
        end
        for j=1,type_index do
          local t = rr_time(lines[i][j])
          if t then
            ttl = t
            lines[i][j] = nil
            break
          end
        end
        if lines[i][1] then name = lines[i][1] end
      end

      type_index = type_index + 2

      if type and class = CLASS_IN and ttl then
        if type == TYPE_A or type == TYPE_AAA then
          ret[#ret+1] = { name=name, type=type, class=class, ttl=ttl, address=lines[i][type_index] }
        elseif type == type_NS then
          ret[#ret+1] = { name=name, type=type, class=class, ttl=ttl, ns=lines[i][type_index] }
        elseif type == TYPE_CNAME then
          ret[#ret+1] = { name=name, type=type, class=class, ttl=ttl, cname=lines[i][type_index] }
        elseif type == TYPE_SOA then
          ret[#ret+1] = { name=name, type=type, class=class, ttl=ttl, mname=lines[i][type_index], rname=lines[i][type_index+1],
                          serial=lines[i][type_index+2], refresh=rr_time(lines[i][type_index+3]), retry=rr_time(lines[i][type_index+4]),
                          expire=rr_time(lines[i][type_index+5]), minimum=rr_time(lines[i][type_index+6]) }
        elseif type == TYPE_PTR then
          ret[#ret+1] = { name=name, type=type, class=class, ttl=ttl, ptr=lines[i][type_index] }
        elseif type == TYPE_MX then
          ret[#ret+1] = { name=name, type=type, class=class, ttl=ttl, preference=lines[i][type_index], exchange=lines[i][type_index+1] }
        elseif type == TYPE_TXT then
          ret[#ret+1] = { name=name, type=type, class=class, ttl=ttl, txt=lines[i][type_index] }
        elseif type == TYPE_SPF then
          ret[#ret+1] = { name=name, type=type, class=class, ttl=ttl, spf=lines[i][type_index] }
        end
      end
    end
  end

  return ret
end

function answer(qname, qtype, qclass, lazy)
  local ret = {}
  for k,v in pairs(parse_file('eder.us', 'eder.us')) do
    ret[#ret+1] = v
  end
  return ret
end
